<?xml version="1.0" encoding="utf-8"?>

<!--
|>
| UNIVERSAL CHEMICAL MARKUP 1-1-1 - SCHEMA UCM RULES VALIDATION SCH [http://www.universalchemicalmarkup.org]
| Updated: 2014-08-22
| License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
| Author: Jan Mokrý (Main corresponding author:
|                    See contact at the project website.)
|>
-->

<schema xmlns="http://purl.oclc.org/dsdl/schematron"
        xmlns:ucm="http://www.universalchemicalmarkup.org"
        xmlns:bibtexml="http://bibtexml.sf.net/"
        xmlns:unitsml="urn:oasis:names:tc:unitsml:schema:xsd:UnitsMLSchema-1.0"
        xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
        xmlns:lf="LocalFunctions"
        queryBinding="xslt2" schemaVersion="ISO19757-3">

<ns prefix="ucm" uri="http://www.universalchemicalmarkup.org"/>
<ns prefix="bibtexml" uri="http://bibtexml.sf.net/"/>
<ns prefix="unitsml" uri="urn:oasis:names:tc:unitsml:schema:xsd:UnitsMLSchema-1.0"/>
<ns prefix="lf" uri="LocalFunctions"/>

<!--PATTERN THAT OUTPUTS INITIAL NEW LINE FOR BETTER READABILITY OF MESSAGES.-->

<pattern>
    <rule context="/*">
        <report test="true()">
        </report>
    </rule>
</pattern>

<!--VALIDATION PATTERNS FOR UCM:*-->

<pattern name="UCM:*@CHARGE">

    <rule context="//ucm:structure[@charge]">
        <assert test="sum(./*/@charge) = ./@charge">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The sum of charges of the node or structure child elements is not equal',
                              'to the charge attribute value on this element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:node[parent::ucm:structure and @idrefs]">
        <let name="ID" value="./@id"/>
        <let name="IDREFS" value="./@idrefs"/>
        <let name="NODE-DEFINITION" value="//ucm:node[parent::ucm:define and @id = $IDREFS]"/>
        <let name="DEFINED-ELECTRON-COUNT"
             value="xs:integer(sum($NODE-DEFINITION/ucm:particle[contains(@type, 'E')]/@counts))"/>
        <let name="SHARED-ELECTRON-COUNT"
             value="xs:integer(sum(//ucm:particle[ucm:share and @idrefs = $ID]/@counts))"/>
        <let name="RELEVANT-SHARE-ELEMENTS" value="//ucm:share[contains(@idrefs, $ID)]"/>
        <let name="NEGATIVE-CHARGE"
             value="lf:COUNT-NODE-NEGATIVE-CHARGE($ID, $DEFINED-ELECTRON-COUNT, $SHARED-ELECTRON-COUNT,
                    $RELEVANT-SHARE-ELEMENTS)"/>
        <let name="POSITIVE-CHARGE" value="sum($NODE-DEFINITION/ucm:particle[@type = 'P']/@counts)"/>
        <let name="TOTAL-CHARGE" value="$POSITIVE-CHARGE - $NEGATIVE-CHARGE"/>
        <let name="CURRENT-CHARGE" value="if (./@charge) then ./@charge else 0"/>
        <assert test="$TOTAL-CHARGE = $CURRENT-CHARGE">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The charge attribute value on this element must be ',
                              $TOTAL-CHARGE, ', because it contains'),
                              concat($POSITIVE-CHARGE, ' proton(s) and ', $NEGATIVE-CHARGE,
                              ' electron(s).')))"/>
        </assert>
    </rule>

    <rule context="//ucm:node[parent::ucm:structure and not(@idrefs)]">
        <let name="ID" value="./@id"/>
        <let name="DEFINED-ELECTRON-COUNT"
             value="xs:integer(sum(./ucm:particle[contains(@type, 'E')]/@counts))"/>
        <let name="SHARED-ELECTRON-COUNT"
             value="xs:integer(sum(//ucm:particle[ucm:share and @idrefs = $ID]/@counts))"/>
        <let name="RELEVANT-SHARE-ELEMENTS" value="//ucm:share[contains(@idrefs, $ID)]"/>
        <let name="NEGATIVE-CHARGE"
             value="lf:COUNT-NODE-NEGATIVE-CHARGE($ID, $DEFINED-ELECTRON-COUNT, $SHARED-ELECTRON-COUNT,
                    $RELEVANT-SHARE-ELEMENTS)"/>
        <let name="POSITIVE-CHARGE" value="sum(./ucm:particle[@type = 'P']/@counts)"/>
        <let name="TOTAL-CHARGE" value="$POSITIVE-CHARGE - $NEGATIVE-CHARGE"/>
        <let name="CURRENT-CHARGE" value="if (./@charge) then ./@charge else 0"/>
        <assert test="$TOTAL-CHARGE = $CURRENT-CHARGE">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The charge attribute value on this element must be ',
                              $TOTAL-CHARGE, ', because it contains'),
                              concat($POSITIVE-CHARGE, ' proton(s) and ', $NEGATIVE-CHARGE,
                              ' electron(s).')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@FRACTIONS">

    <rule context="//ucm:*[@fractions]">
        <assert test="sum(for $i in tokenize(./@fractions,'\s+') return xs:decimal($i)) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The sum of all numbers in the fractions attribute on this element is not',
                              'equal to one.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@FORMAT-1">

    <rule context="//ucm:*[matches(@format, '^(UNITSML|BIBTEXML)$')]">
        <assert test="local-name() = 'define'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The format attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on define elements.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[@format and not(matches(@format, '^(UCM|UNITSML|BIBTEXML)$'))]">
        <let name="INCHI-FORMULA-REGEX" value="string('(/([0-9]*[A-IK-PR-Z][a-ik-pr-vy]?[0-9]*\.?)+)?')"/>
        <let name="SMILES-CHARS-REGEX" value="string('0-9A-IK-PR-Za-ik-pr-vy*@:.=#$\\/%()\[\]+\-')"/>
        <let name="ENABLED-CONTENT-REGEX"
             value="if (@format = 'IUPAC-PREFERRED-NAME-U'
                    or @format = 'IUPAC-GENERAL-NAME'
                    or @format = 'CA-INDEX-NAME') then '^\s*(\S| )+\s*$'
                    else (if (@format = 'CAS-RN-U') then '^\s*[0-9]{1,7}-[0-9]{2}-[0-9]\s*$'
                    else (if (@format = 'REAXYS-RN-U'
                    or @format = 'CHEMSPIDER-ID-U'
                    or @format = 'PUBCHEM-CID-U'
                    or @format = 'PUBCHEM-SID') then '^\s*[0-9]+\s*$'
                    else (if (@format = 'INCHI')
                    then concat('^\s*1', $INCHI-FORMULA-REGEX,
                    '(/[bcfhimopqrst][0-9A-IK-PR-Za-ik-pr-vy.,;()?*+\-]*)*\s*$')
                    else (if (@format = 'INCHI-KEY') then '^\s*[A-Z]{14}-[A-Z]{8}NA-[A-Z]\s*$'
                    else (if (@format = 'S-INCHI-U')
                    then concat('^\s*1S', $INCHI-FORMULA-REGEX,
                    '(/[bchimpqst][0-9DHThm.,;()?*+\-]*)*\s*$')
                    else (if (@format = 'S-INCHI-KEY') then '^\s*[A-Z]{14}-[A-Z]{8}SA-[A-Z]\s*$'
                    else (if (@format = 'SMILES')
                    then concat('^\s*[', $SMILES-CHARS-REGEX, ']+\s*$')
                    else (if (@format = 'SMARTS')
                    then concat('^\s*[', $SMILES-CHARS-REGEX, 'x;,~?!&amp;]+\s*$')
                    else (if (@format = 'SLN')
                    then '^\s*[0-9A-Za-z*@;:.,~=#(){}\[\]+\-\\|!&amp;&quot;&lt;&gt;]+\s*$'
                    else '^$')))))))))"/>
        <let name="NOT-ENABLED-CONTENT-REGEX" value="if (@format = 'SLN') then '^\s*.*-&gt;.*\s*$' else '^$'"/>
        <assert test="local-name() = 'structure'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The format attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on structure elements.'))"/>
        </assert>
        <assert test="matches(./text(), $ENABLED-CONTENT-REGEX)
                      and not(matches(./text(), $NOT-ENABLED-CONTENT-REGEX))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The content inside this element does not match the regular expresion',
                              concat('pattern for the given format attribute value &quot;', ./@format,
                              '&quot;.')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@FORMAT-2">

    <rule context="//ucm:*[@format = 'UCM']">
        <assert test="count(./*) = count(./ucm:*)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('On this element the format attribute has the value &quot;UCM&quot;, but',
                              'not all child elements inside the element are from UCM namespace.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[@format = 'UNITSML']">
        <assert test="count(./*) = count(./unitsml:*)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('On this element the format attribute has the value &quot;UNITSML&quot;, but',
                              'not all child elements inside the element are from UnitsML namespace.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[@format = 'BIBTEXML']">
        <assert test="count(./*) = count(./bibtexml:*)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('On this element the format attribute has the value &quot;BIBTEXML&quot;, but',
                              'not all child elements inside the element are from BibTeXML namespace.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@FORMAT-3">

    <rule context="//ucm:*[@type = 'SBST']">
        <assert test="./@format = 'UCM'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The type attribute on this element has the value &quot;SBST&quot;, thus',
                              'the value of the format attribute must be &quot;UCM&quot;, but instead',
                              concat('it is &quot;', ./@format, '&quot;.')))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[@type = 'STQR']">
        <assert test="not(./@format = 'UCM')">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The type attribute on this element has the value &quot;STQR&quot;, thus',
                              'the value of the format attribute cannot be &quot;UCM&quot;.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[@type = 'STID']">
        <assert test="not(matches(@format, '^(UCM|SMILES|SMARTS|SLN)$'))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The type attribute on this element has the value &quot;STID&quot;, thus',
                              concat('the value of the format attribute cannot be &quot;',
                              ./@format, '&quot;.')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@IDREFS">

    <rule context="//ucm:*[@idrefs]">
        <let name="IDREFS" value="tokenize(./@idrefs,'\s+')"/>
        <assert test="deep-equal(lf:SORT(distinct-values($IDREFS)), lf:SORT($IDREFS))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference is repeated more than once in the idrefs',
                              'attribute on this element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:*@TYPE">

    <rule context="//ucm:*[matches(@type, '^(P|N|E|N?BE)$')]">
        <assert test="local-name() = 'particle'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on particle elements.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[matches(@type, '^(PR|CN|ER)$')]">
        <assert test="local-name() = 'property'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on property elements.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[matches(@type, '^(ST|STQR)$')]">
        <assert test="local-name() = 'structure' and parent::ucm:*[local-name() = 'ucm']">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on structure elements that are the childs of an ucm',
                              'element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:*[matches(@type, '^(SBST|STID)$')]">
        <assert test="local-name() = 'structure' and parent::ucm:*[local-name() = 'structure']">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              'which is enabled only on structure elements that are the childs of another',
                              'structure element.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:BOND-->

<pattern name="UCM:BOND-1">

    <rule context="//ucm:bond[ucm:particle]">
        <let name="ID-REGEX" value="concat('(^| )', ./@id, '( |$)')"/>
        <let name="BOND-BONDING-ELECTRONS" value="sum(./ucm:particle[@type = 'BE']/@counts)"/>
        <let name="REQUIRED-BOND-BONDING-ELECTRONS"
             value="sum(//ucm:particle[matches(@idrefs, $ID-REGEX) and @type = 'BE']/@counts)"/>
        <assert test="$BOND-BONDING-ELECTRONS >= $REQUIRED-BOND-BONDING-ELECTRONS">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('This element contains ', $BOND-BONDING-ELECTRONS,
                              ' bonding electron(s), but all bond elements that refer'),
                              concat('to this element require ', $REQUIRED-BOND-BONDING-ELECTRONS,
                              ' bonding electron(s).')))"/>
        </assert>
    </rule>

    <rule context="//ucm:bond[not(ucm:particle)]">
        <let name="ID-REGEX" value="concat('(^| )', ./@id, '( |$)')"/>
        <let name="BOND-BONDING-ELECTRONS" value="if (@order = 'S') then 2
                                                  else (if (@order = 'D') then 4
                                                  else (if (@order = 'T') then 6
                                                  else (if (@order = 'Q') then 8
                                                  else 0)))"/>
        <let name="REQUIRED-BOND-BONDING-ELECTRONS"
             value="sum(//ucm:particle[matches(@idrefs, $ID-REGEX) and @type = 'BE']/@counts)"/>
        <assert test="$BOND-BONDING-ELECTRONS >= $REQUIRED-BOND-BONDING-ELECTRONS">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('This element contains ', $BOND-BONDING-ELECTRONS,
                              ' bonding electron(s), but all bond elements that refer'),
                              concat('to this element require ', $REQUIRED-BOND-BONDING-ELECTRONS,
                              ' bonding electron(s).')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND-2">

    <rule context="//ucm:bond[not(@idrefs)]">
        <assert test="count(./ucm:join) >= 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which does not have an idrefs attribute, is missing one or more',
                              'mandatory join child element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:bond[@idrefs]">
        <assert test="not(./ucm:join)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which has the idrefs attribute, cannot contain any join child',
                              'elements.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND-3">

    <rule context="//ucm:bond[matches(@order, '^(P?S|P?D|P?T|P?Q)$')]">
        <assert test="./@idrefs">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The order attribute on this element has the value &quot;',
                              ./@order, '&quot;,'),
                              'thus the context with the idrefs attribute is mandatory, but the idrefs',
                              'attribute is missing on this element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND@IDREFS-1">

    <rule context="//ucm:bond[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) = 2">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There are not two id references in the idrefs attribute on this element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', ., //(ucm:node | ucm:point)[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node or point element inside a structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND@IDREFS-2">

    <rule context="//ucm:bond[not(ucm:particle) and @idrefs]">
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:node[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element does not contain any particle child element, but at least one',
                              'id reference in its idrefs attribute does not refer to the node element',
                              'inside a structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND@ORDER-1">

    <rule context="//ucm:bond[matches(@order, '^(PS|PD|PT|PQ|A|DL)$')]">
        <assert test="count(./ucm:particle[@type = 'BE']) >= 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The order attribute on this element has the value &quot;',
                              ./@order, '&quot;,'),
                              'but electrons participating in the bond are not specified using one or more',
                              'particle elements inside this element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:bond[matches(@order, '^(I|H|DIP)$')]">
        <assert test="count(./ucm:particle) = 0">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The order attribute on this element has the value &quot;',
                              ./@order, '&quot;,'),
                              'thus particle child elements for this element are not enabled, but there',
                              'is at least one.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:BOND@ORDER-2">

    <rule context="//ucm:bond[ucm:particle and matches(@order, '^(P?S|P?D|P?T|P?Q)$')]">
        <let name="BOND-BONDING-ELECTRONS" value="sum(./ucm:particle[@type = 'BE']/@counts)"/>
        <let name="REQUIRED-BOND-BONDING-ELECTRONS" value="if (@order = 'PS') then 1
                                                  else (if (@order = 'S') then 2
                                                  else (if (@order = 'PD') then 3
                                                  else (if (@order = 'D') then 4
                                                  else (if (@order = 'PT') then 5
                                                  else (if (@order = 'T') then 6
                                                  else (if (@order = 'PQ') then 7
                                                  else (if (@order = 'Q') then 8
                                                  else 0)))))))"/>
        <assert test="$BOND-BONDING-ELECTRONS = $REQUIRED-BOND-BONDING-ELECTRONS">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('This element contains ', $BOND-BONDING-ELECTRONS,
                              ' bonding electron(s), but the order attribute on this'),
                              concat('element has the value &quot;', ./@order, '&quot;, which requires ',
                              $REQUIRED-BOND-BONDING-ELECTRONS, ' bonding electron(s).')))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:DESCRIPTION-->

<pattern name="UCM:DESCRIPTION">

    <rule context="//ucm:description[parent::ucm:define]">
        <assert test="./@id">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is no id attribute on this element, but on the description element',
                              'that is inside a define element the id attribute is mandatory.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:DESCRIPTION@IDREFS">

    <rule context="//ucm:description[@idrefs]">
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:description[parent::ucm:define and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the description element inside a define element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:DESCRIPTION@LITREFS">

    <rule context="//ucm:description[@litrefs]">
        <let name="LITREFS" value="tokenize(./@litrefs,'\s+')"/>
        <assert test="deep-equal(lf:SORT(distinct-values($LITREFS)), lf:SORT($LITREFS))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference is repeated more than once in the litrefs',
                              'attribute on this element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('litrefs', ., //bibtexml:entry[@id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the litrefs attribute on this element does not',
                              'refer to the literature reference defined using BibTeXML inside a define',
                              'element.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:JOIN-->

<pattern name="UCM:JOIN@IDREFS-1">

    <rule context="//ucm:join[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) >= 2">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is less than two id references in the idrefs attribute on this',
                              'element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', ., //(ucm:node | ucm:point)[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node or point element inside a structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:JOIN@IDREFS-2">

    <rule context="//ucm:join[parent::ucm:bond and not(./parent::ucm:bond/ucm:particle)]">
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:node[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The parent bond element of this element does not contain any particle child',
                              'element, but at least one id reference in the idrefs attribute on this element',
                              'does not refer to the node element inside a structure element.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:NODE-->

<pattern name="UCM:NODE-1">

    <rule context="//ucm:node[not(@idrefs)]">
        <assert test="count(./ucm:particle) >= 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which does not have an idrefs attribute, is missing one or more',
                              'mandatory particle child element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:node[parent::ucm:strucure and @idrefs]">
        <assert test="not(./ucm:particle)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which is inside a strucure element and has the idrefs',
                              'attribute, cannot contain any particle child elements.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:NODE-2">

    <rule context="//ucm:node[parent::ucm:structure and @idrefs]">
        <let name="IDREFS" value="./@idrefs"/>
        <let name="NODE-BONDING-ELECTRONS"
             value="sum(//ucm:node[parent::ucm:define and @id = $IDREFS]/ucm:particle[@type = 'BE']/@counts)"/>
        <let name="REQUIRED-NODE-BONDING-ELECTRONS"
             value="lf:COUNT-NODE-BONDING-ELECTRONS(.,
                    //ucm:bond[not(ucm:particle)],
                    //ucm:particle[@type = 'BE'])"/>
        <assert test="$NODE-BONDING-ELECTRONS = $REQUIRED-NODE-BONDING-ELECTRONS">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The node definition of this element contains ',
                              $NODE-BONDING-ELECTRONS, ' bonding electron(s), but all'),
                              concat('bonds of this element require ',
                              $REQUIRED-NODE-BONDING-ELECTRONS, ' bonding electron(s).')))"/>
        </assert>
    </rule>

    <rule context="//ucm:node[parent::ucm:structure and not(@idrefs)]">
        <let name="NODE-BONDING-ELECTRONS" value="sum(./ucm:particle[@type = 'BE']/@counts)"/>
        <let name="REQUIRED-NODE-BONDING-ELECTRONS"
             value="lf:COUNT-NODE-BONDING-ELECTRONS(.,
                    //ucm:bond[not(ucm:particle)],
                    //ucm:particle[@type = 'BE'])"/>
        <assert test="$NODE-BONDING-ELECTRONS = $REQUIRED-NODE-BONDING-ELECTRONS">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('This element contains ', $NODE-BONDING-ELECTRONS,
                              ' bonding electron(s), but all bonds of this element'),
                              concat('require ', $REQUIRED-NODE-BONDING-ELECTRONS,
                              ' bonding electron(s).')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:NODE-3">

    <rule context="//ucm:node[parent::ucm:define]">
        <assert test="not(./@charge)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is a charge attribute on this element, but on the node element that',
                              'is inside a define element the charge attribute must be omitted.'))"/>
        </assert>
        <assert test="not(./@x) or not(./@y) or not(./@z)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element contains coordinates in 3-dimensional space (at least one of',
                              'the x, y or z attributes), but on the node element that is inside a define',
                              'element the attributes for coordinates must be omitted.'))"/>
        </assert>
        <assert test="not(./ucm:stereo)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element contains a stereo child element, but on the node element that',
                              'is inside a define element the stereo child element must be omitted.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:NODE@IDREFS">

    <rule context="//ucm:node[parent::ucm:define]">
        <assert test="not(./@idrefs)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is an idrefs attribute on this element, but on the node element that',
                              'is inside a define element the idrefs attribute must be omitted.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:node[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is not one id reference in the idrefs attribute on this element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:node[parent::ucm:define and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node element inside a define element.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:PARTICLE-->

<pattern name="UCM:PARTICLE">

    <rule context="//ucm:particle[parent::ucm:bond]">
        <assert test="./@idrefs">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is no idrefs attribute on this element, but on the particle element',
                              'that is inside a bond element the idrefs attribute is mandatory.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:particle[parent::ucm:node]">
        <assert test="not(./@idrefs)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is an idrefs attribute on this element, but on the particle element',
                              'that is inside a node element the idrefs attribute must be omitted.'))"/>
        </assert>
        <assert test="not(./ucm:share)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which is inside a node element, cannot contain any share child',
                              'elements.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PARTICLE@COUNTS">

    <rule context="//ucm:particle[count(tokenize(./@counts,'\s+')) > 1]">
        <assert test="./@type = 'N'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The counts attribute on this element has more than one number in its value,',
                              'thus the value of the type attribute on this element must be &quot;N&quot;,',
                              concat('but instead it is &quot;', ./@type, '&quot;.')))"/>
        </assert>
        <assert test="count(tokenize(./@counts,'\s+')) = count(tokenize(./@fractions,'\s+'))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('On this element the occurrence ratio of each count in the counts attribute',
                              'is not specified properly using the fractions attribute.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:particle[count(tokenize(./@counts,'\s+')) = 1]">
        <assert test="not(./@fractions)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is a fractions attribute on this element, but on the particle element,',
                              'with one number in its counts attribute the fractions attribute must be',
                              'omitted.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PARTICLE@IDREFS">

    <rule context="//ucm:particle[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is not one id reference in the idrefs attribute on this element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', .,
                      (//ucm:bond[@id] | //ucm:structure[@id] | //ucm:node[parent::ucm:structure and @id]))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the bond or structure element or to the node element inside',
                              'the structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PARTICLE@TYPE-1">

    <rule context="//ucm:particle[parent::ucm:bond and @type]">
        <assert test="./@type = 'BE'">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;, but'),
                              'only the &quot;BE&quot; value of the type attribute is enabled on the particle',
                              'element that is inside the bond element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:particle[parent::ucm:node and @type]">
        <let name="TYPE" value="./@type"/>
        <assert test="count(./parent::ucm:node/ucm:particle[@type = $TYPE]) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(./parent::ucm:node,
                              ('This element cannot contain more than one particle element with the type',
                              concat('attribute that has the value &quot;', ./@type, '&quot;.')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PARTICLE@TYPE-2">

    <rule context="//ucm:particle[parent::ucm:node and @type = 'NBE']">
        <assert test="count(./parent::ucm:node/ucm:particle[@type = 'BE']) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(./parent::ucm:node,
                              ('This element must contain the particle element with the type attribute that',
                              'has the value &quot;BE&quot;, because it contains the particle element that',
                              concat('has the type attribute with the value &quot;', ./@type, '&quot;.')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PARTICLE@TYPE-3">

    <rule context="//ucm:particle[parent::ucm:node and (@type = 'BE' or @type = 'NBE')]">
        <assert test="count(./parent::ucm:node/ucm:particle[@type = 'E']) = 0">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(./parent::ucm:node,
                              ('This element cannot contain the particle element with the type attribute that',
                              'has the value &quot;E&quot;, because it contains the particle element that has',
                              concat('the type attribute with the value &quot;', ./@type, '&quot;.')))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:PROPERTY-->

<pattern name="UCM:PROPERTY">

    <rule context="//ucm:property[not(@idrefs)]">
        <assert test="./ucm:values">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which does not have an idrefs attribute, is missing one',
                              'mandatory values child element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@idrefs]">
        <assert test="not(./ucm:property) and not(./ucm:values)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('This element, which has the idrefs attribute, cannot contain any',
                              'property or values child elements.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PROPERTY@IDREFS">

    <rule context="//ucm:property[parent::ucm:define]">
        <assert test="not(./@idrefs)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is an idrefs attribute on this element, but on the property element',
                              'that is inside a define element the idrefs attribute must be omitted.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is not one id reference in the idrefs attribute on this element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:property[parent::ucm:define and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the property element inside a define element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PROPERTY@QUANTITY-1">

    <rule context="//ucm:property[@quantity]">
        <let name="QUANTITY" value="substring-after(./@quantity, '#')"/>
        <assert test="count(//unitsml:Quantity[@xml:id = $QUANTITY]) = 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The quantity attribute value on this element does not refer to the unique',
                              'xml:id of a quantity defined using UnitsML inside a define element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PROPERTY@QUANTITY-2">

    <rule context="//ucm:property[not(@idrefs)]">
        <assert test="./@quantity">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is no quantity attribute on this element, but on the property element',
                              'without the idrefs attribute the quantity attribute is mandatory.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@idrefs]">
        <assert test="not(./@quantity)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is a quantity attribute on this element, but on the property element',
                              'that has the idrefs attribute the quantity attribute must be omitted.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PROPERTY@TYPE-1">

    <rule context="//ucm:property[@type = 'PR']">
        <assert test="parent::ucm:*[local-name() = 'bond' or local-name() = 'node'
                      or local-name() = 'particle' or local-name() = 'point'
                      or local-name() = 'structure' or local-name() = 'define']">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              concat('thus this element must not be a child of the ',
                              name(parent::*), ' element.')))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@type = 'CN']">
        <assert test="parent::ucm:*[local-name() = 'property' or local-name() = 'define']">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              concat('thus this element must not be a child of the ',
                              name(parent::*), ' element.')))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@type = 'ER']">
        <assert test="parent::ucm:*[local-name() = 'property']">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              (concat('The type attribute on this element has the value &quot;',
                              ./@type, '&quot;,'),
                              concat('thus this element must not be a child of the ',
                              name(parent::*), ' element.')))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:PROPERTY@TYPE-2">

    <rule context="//ucm:property[not(@idrefs)]">
        <assert test="./@type">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is no type attribute on this element, but on the property element',
                              'without the idrefs attribute, the type attribute is mandatory.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:property[@idrefs]">
        <assert test="not(./@type)">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There is a type attribute on this element, but on the property element that',
                              'has the idrefs attribute the type attribute must be omitted.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:SHARE-->

<pattern name="UCM:SHARE@IDREFS">

    <rule context="//ucm:share[@idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) > 1">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There are not two or more id references in the idrefs attribute on this',
                              'element.'))"/>
        </assert>
        <assert test="lf:CHECK-REFS('idrefs', ., //ucm:node[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node element inside a structure element.'))"/>
        </assert>
        <assert test="count(tokenize(./@idrefs,'\s+')) = count(tokenize(./@fractions,'\s+'))">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('On this element the sharing ratio of each id reference in the idrefs',
                              'attribute is not specified properly using the fractions attribute.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION PATTERNS FOR UCM:STEREO-->

<pattern name="UCM:STEREO@IDREFS-1">

    <rule context="//ucm:stereo[@idrefs]">
        <assert test="lf:CHECK-REFS('idrefs', ., //(ucm:node | ucm:point)[parent::ucm:structure and @id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node or point element inside a structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:STEREO@IDREFS-2">

    <rule context="//ucm:stereo[parent::ucm:bond and @idrefs]">
        <assert test="count(tokenize(./@idrefs,'\s+')) = 4">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There are not four id references in the idrefs attribute on this element',
                              'although it is inside the bond element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:stereo[parent::ucm:node and @idrefs]">
        <let name="IDREFS-COUNT" value="count(tokenize(./@idrefs,'\s+'))"/>
        <assert test="$IDREFS-COUNT = 4 or $IDREFS-COUNT = 5 or $IDREFS-COUNT = 7">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There are not four, five or seven id references in the idrefs attribute',
                              'on this element although it is inside the node element.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:stereo[parent::ucm:structure and @idrefs]">
        <let name="IDREFS-COUNT" value="count(tokenize(./@idrefs,'\s+'))"/>
        <assert test="$IDREFS-COUNT = 5 or $IDREFS-COUNT = 6 or $IDREFS-COUNT = 7">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('There are not five, six or seven id references in the idrefs attribute',
                              'on this element although it is inside the structure element.'))"/>
        </assert>
    </rule>

</pattern>

<pattern name="UCM:STEREO@IDREFS-3">

    <rule context="//ucm:stereo[parent::ucm:node and @idrefs and count(tokenize(@idrefs,'\s+')) > 4]">
        <let name="IDREFS" value="tokenize(./@idrefs,'\s+')"/>
        <assert test="$IDREFS[1] = ./parent::ucm:node/@id">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('The first id reference in the idrefs attribute on this element does not refer',
                              'to the parent node element although this element is inside the node element',
                              'and has more than four id references in the idrefs attribute.'))"/>
        </assert>
    </rule>

    <rule context="//ucm:stereo[parent::ucm:structure and @idrefs and count(tokenize(@idrefs,'\s+')) > 5]">
        <assert test="lf:CHECK-REFS('idrefs', ., ./parent::ucm:structure/descendant::ucm:node[@id])">
            <value-of select="lf:OUTPUT-ERROR-MESSAGE(.,
                              ('At least one id reference in the idrefs attribute on this element does not',
                              'refer to the node element that is a descendant of the parent structure element',
                              'although this element is inside the structure element and has more than five',
                              'id references in the idrefs attribute.'))"/>
        </assert>
    </rule>

</pattern>

<!--VALIDATION LOCAL FUNCTIONS-->

<xsl:function name="lf:CHECK-REFS" as="xs:boolean">

    <xsl:param name="REFS-ATTRIBUTE-NAME" as="xs:string"/>
    <xsl:param name="ELEMENT-WITH-REFS" as="node()"/>
    <xsl:param name="RELEVANT-ELEMENTS-WITH-ID" as="node()*"/>

    <xsl:variable name="RELEVANT-ID-LIST">
        <xsl:for-each select="$RELEVANT-ELEMENTS-WITH-ID">
            <xsl:value-of select="./@id"/>
            <xsl:if test="position() != last()">
                <xsl:text> </xsl:text>
            </xsl:if>
        </xsl:for-each>
    </xsl:variable>

    <xsl:variable name="CHECK-RESULT">
        <xsl:for-each select="tokenize($ELEMENT-WITH-REFS/@*[name() = $REFS-ATTRIBUTE-NAME],'\s+')">
            <xsl:if test="not(matches($RELEVANT-ID-LIST, concat('(^| )', ., '( |$)')))">
                <xsl:text>F</xsl:text>
            </xsl:if>
        </xsl:for-each>
    </xsl:variable>

    <xsl:choose>
        <xsl:when test="contains($CHECK-RESULT, 'F')">
            <xsl:value-of select="false()"/>
        </xsl:when>
        <xsl:otherwise>
            <xsl:value-of select="true()"/>
        </xsl:otherwise>
    </xsl:choose>

</xsl:function>

<xsl:function name="lf:COUNT-NODE-NEGATIVE-CHARGE" as="xs:decimal">

    <xsl:param name="NODE-ID" as="xs:string"/>
    <xsl:param name="DEFINED-ELECTRON-COUNT" as="xs:integer"/>
    <xsl:param name="SHARED-ELECTRON-COUNT" as="xs:integer"/>
    <xsl:param name="RELEVANT-SHARE-ELEMENTS" as="node()*"/>

    <xsl:variable name="FRACTIONS-OF-SHARED-ELECTRON-COUNTS">
        <xsl:for-each select="$RELEVANT-SHARE-ELEMENTS">
            <xsl:value-of select="xs:decimal(subsequence(tokenize(./@fractions, '\s+'),
                                  index-of(tokenize(./@idrefs, '\s+'), $NODE-ID), 1))
                                  * xs:decimal(./parent::ucm:particle/@counts)"/>
            <xsl:if test="position() != last()">
                <xsl:text> </xsl:text>
            </xsl:if>
        </xsl:for-each>
    </xsl:variable>

    <xsl:value-of select="$DEFINED-ELECTRON-COUNT - $SHARED-ELECTRON-COUNT + sum(
                          for $i in tokenize($FRACTIONS-OF-SHARED-ELECTRON-COUNTS,' ')
                          return xs:decimal($i))"/>

</xsl:function>

<xsl:function name="lf:COUNT-NODE-BONDING-ELECTRONS" as="xs:integer">

    <xsl:param name="NODE-ELEMENT" as="node()"/>
    <xsl:param name="RELEVANT-BOND-ELEMENTS" as="node()*"/>
    <xsl:param name="RELEVANT-PARTICLE-ELEMENTS" as="node()*"/>

    <xsl:variable name="PARENT-STRUCTURE-ELEMENT" select="$NODE-ELEMENT/parent::ucm:structure"/>
    <xsl:variable name="PARENT-STRUCTURE-ELEMENT-ID-REGEX"
                  select="concat('(^| )', $PARENT-STRUCTURE-ELEMENT/@id, '( |$)')"/>
    <xsl:variable name="NODE-ELEMENT-ID-REGEX" select="concat('(^| )', $NODE-ELEMENT/@id, '( |$)')"/>
    <xsl:variable name="NODE-COUNT" select="count($PARENT-STRUCTURE-ELEMENT/ucm:node)"/>

    <xsl:variable name="SINGLE-BONDS-ELECTRON-COUNT"
                  select="count($RELEVANT-BOND-ELEMENTS[matches(@idrefs, $NODE-ELEMENT-ID-REGEX)
                          and @order = 'S'])"/>
    <xsl:variable name="DOUBLE-BONDS-ELECTRON-COUNT"
                  select="count($RELEVANT-BOND-ELEMENTS[matches(@idrefs, $NODE-ELEMENT-ID-REGEX)
                          and @order = 'D']) * 2"/>
    <xsl:variable name="TRIPLE-BONDS-ELECTRON-COUNT"
                  select="count($RELEVANT-BOND-ELEMENTS[matches(@idrefs, $NODE-ELEMENT-ID-REGEX)
                          and @order = 'T']) * 3"/>
    <xsl:variable name="QUADRUPLE-BONDS-ELECTRON-COUNT"
                  select="count($RELEVANT-BOND-ELEMENTS[matches(@idrefs, $NODE-ELEMENT-ID-REGEX)
                          and @order = 'Q']) * 4"/>
    <xsl:variable name="OTHER-BONDS-ELECTRON-COUNT"
                  select="sum($RELEVANT-PARTICLE-ELEMENTS[matches(@idrefs, $NODE-ELEMENT-ID-REGEX)]/@counts)
                          + (sum($RELEVANT-PARTICLE-ELEMENTS[matches(@idrefs,
                          $PARENT-STRUCTURE-ELEMENT-ID-REGEX)]/@counts) div $NODE-COUNT)"/>

    <xsl:value-of select="$SINGLE-BONDS-ELECTRON-COUNT + $DOUBLE-BONDS-ELECTRON-COUNT
                          + $TRIPLE-BONDS-ELECTRON-COUNT + $QUADRUPLE-BONDS-ELECTRON-COUNT
                          + $OTHER-BONDS-ELECTRON-COUNT"/>

</xsl:function>

<xsl:function name="lf:OUTPUT-ERROR-MESSAGE" as="xs:string">

    <xsl:param name="NODE" as="node()"/>
    <xsl:param name="MESSAGE-LINES" as="xs:string*"/>

    <xsl:value-of select="concat('&#10;Error in the &quot;', lf:OUTPUT-PATH-TO-NODE($NODE),
                          '&quot; element:&#10;  ', string-join($MESSAGE-LINES, '&#10;  '))"/>

</xsl:function>

<xsl:function name="lf:OUTPUT-INDEX-OF-NODE" as="xs:integer">
    <!--Adapted from the souce code released under the GNU LGPL 3 license at
        "http://www.xsltfunctions.com/xsl/functx_index-of-node.html".-->

    <xsl:param name="NODES" as="node()*"/>
    <xsl:param name="CURRENT-NODE" as="node()"/>

    <xsl:value-of select="for $INDEX in (1 to count($NODES))
                          return $INDEX[$NODES[$INDEX] is $CURRENT-NODE]"/>

</xsl:function>

<xsl:function name="lf:OUTPUT-PATH-TO-NODE" as="xs:string">
    <!--Adapted from the souce code released under the GNU LGPL 3 license at
        "http://www.xsltfunctions.com/xsl/functx_path-to-node-with-pos.html".-->

    <xsl:param name="NODE" as="node()"/>

    <xsl:variable name="NAMES" as="xs:string*">
        <xsl:for-each select="$NODE/ancestor-or-self::*">
            <xsl:variable name="NODE-NAME" select="name(.)"/>
            <xsl:variable name="SAME-NAME-NODES" select="./parent::*/*[name() = $NODE-NAME]"/>
            <xsl:sequence select="concat($NODE-NAME,
                                  if (count($SAME-NAME-NODES) > 1)
                                  then concat('[', lf:OUTPUT-INDEX-OF-NODE($SAME-NAME-NODES, .), ']')
                                  else '')"/>
        </xsl:for-each>
    </xsl:variable>

    <xsl:sequence select="string-join($NAMES, '/')"/>

</xsl:function>

<xsl:function name="lf:SORT" as="item()*">

    <xsl:param name="SEQUENCE" as="item()*"/>

    <xsl:for-each select="$SEQUENCE">
        <xsl:sort select="."/>
        <xsl:copy-of select="."/>
    </xsl:for-each>

</xsl:function>

</schema>
